@@ -129,11 +129,12 @@ static void preserve_perms(int fd_in, int fd_out)
 
 	if (unlikely(fstat(fd_in, &st)))
 		fatal("Failed to fstat input file\n");
-	if (fchmod(fd_out, (st.st_mode & 0777)))
+	if (unlikely(fchmod(fd_out, (st.st_mode & 0777))))
 		print_err("Warning, unable to set permissions on %s\n", control.outfile);
 
 	/* chown fail is not fatal */
-	fchown(fd_out, st.st_uid, st.st_gid);
+	if (unlikely(fchown(fd_out, st.st_uid, st.st_gid)))
+		print_err("Warning, unable to set owner on %s\n", control.outfile);
 }
 
 /* Open a temporary outputfile to emulate stdout */
@@ -375,15 +376,18 @@ static void get_fileinfo(void)
 	/* Version < 0.4 had different file format */
 	if (control.major_version == 0 && control.minor_version < 4)
 		seekspot = 50;
-	else
+	else if (control.major_version == 0 && control.minor_version == 4)
 		seekspot = 74;
+	else
+		seekspot = 75;
 	if (unlikely(lseek(fd_in, seekspot, SEEK_SET) == -1))
 		fatal("Failed to lseek in get_fileinfo: %s\n", strerror(errno));
 
 	/* Read the compression type of the first block. It's possible that
 	   not all blocks are compressed so this may not be accurate.
 	 */
-	read(fd_in, &ctype, 1);
+	if (unlikely(read(fd_in, &ctype, 1) != 1))
+		fatal("Failed to read in get_fileinfo\n");
 
 	cratio = (long double)expected_size / (long double)infile_size;
 	
@@ -401,6 +405,8 @@ static void get_fileinfo(void)
 		print_output("rzip + gzip\n");
 	else if (ctype == CTYPE_ZPAQ)
 		print_output("rzip + zpaq\n");
+	else
+		print_output("Dunno wtf\n");
 	print_output("Decompressed file size: %llu\n", expected_size);
 	print_output("Compressed file size: %llu\n", infile_size);
 	print_output("Compression ratio: %.3Lf\n", cratio);
@@ -702,13 +708,13 @@ int main(int argc, char *argv[])
 
 	/* OK, if verbosity set, print summary of options selected */
 	if (!INFO) {
-		print_verbose("The following options are in effect for this %s.\n",
-			DECOMPRESS ? "DECOMPRESSION" : "COMPRESSION");
-		if (LZMA_COMPRESS)
-			print_verbose("Threading is %s. Number of CPUs detected: %d\n", control.threads > 1? "ENABLED" : "DISABLED",
-				control.threads);
+		if (!TEST_ONLY)
+			print_verbose("The following options are in effect for this %s.\n",
+				      DECOMPRESS ? "DECOMPRESSION" : "COMPRESSION");
+		print_verbose("Threading is %s. Number of CPUs detected: %d\n", control.threads > 1? "ENABLED" : "DISABLED",
+			      control.threads);
 		print_verbose("Detected %lld bytes ram\n", control.ramsize);
-		print_verbose("Comrpession level %d\n", control.compression_level);
+		print_verbose("Compression level %d\n", control.compression_level);
 		print_verbose("Nice Value: %d\n", control.nice_val);
 		print_verbose("Show Progress\n");
 		print_maxverbose("Max ");
@@ -725,7 +731,7 @@ int main(int argc, char *argv[])
 			print_verbose("Test file integrity\n");
 
 		/* show compression options */
-		if (!DECOMPRESS) {
+		if (!DECOMPRESS && !TEST_ONLY) {
 			print_verbose("Compression mode is: ");
 			if (LZMA_COMPRESS)
 				print_verbose("LZMA. LZO Test Compression Threshold: %.f\n",
@@ -749,8 +755,19 @@ int main(int argc, char *argv[])
 		}
 	}
 
-	if (unlikely(setpriority(PRIO_PROCESS, 0, control.nice_val) == -1))
-		print_err("Warning, unable to set nice value\n");
+	/* Decrease usable ram size on 32 bits due to kernel/userspace split */
+	if (BITS32)
+		control.ramsize = MAX(control.ramsize - 900000000ll, 900000000ll);
+
+	/* Set the main nice value to half that of the backend threads since
+	 * the rzip stage is usually the rate limiting step */
+	if (control.nice_val > 0) {
+		if (unlikely(setpriority(PRIO_PROCESS, 0, control.nice_val / 2) == -1))
+			print_err("Warning, unable to set nice value\n");
+	} else {
+		if (unlikely(setpriority(PRIO_PROCESS, 0, control.nice_val) == -1))
+			print_err("Warning, unable to set nice value\n");
+	}
 
 	/* One extra iteration for the case of no parameters means we will default to stdin/out */
 	for (i = 0; i <= argc; i++) {
