@@ -99,22 +99,29 @@ struct sliding_buffer {
 	i64 size_high;	/* "" high "" */
 	i64 high_length;/* How big the high buffer should be */
 	int fd;		/* The fd of the mmap */
+	int low_top;	/* The low buffer has hit the top */
 } sb;	/* Sliding buffer */
 
+static void round_to_page(i64 *size)
+{
+	*size -= *size % control.page_size;
+	if (unlikely(!*size))
+		*size = control.page_size;
+}
+
 static void remap_low_sb(void)
 {
-	static int top = 0;
 	i64 new_offset;
 
-	if (top)
+	if (sb.low_top)
 		return;
 	new_offset = sb.offset_search;
 	if (new_offset + sb.size_low > sb.orig_size) {
 		new_offset = sb.orig_size - sb.size_low;
-		top = 1;
+		sb.low_top = 1;
 	}
-	new_offset -= new_offset % control.page_size; /* Round to page size */
-	print_maxverbose("Sliding main buffer\n");
+	round_to_page(&new_offset);
+	print_maxverbose("Sliding main buffer to offset %lld\n", new_offset);
 	if (unlikely(munmap(sb.buf_low, sb.size_low)))
 		fatal("Failed to munmap in remap_low_sb\n");
 	sb.offset_low = new_offset;
@@ -159,20 +166,21 @@ static uchar *get_sb(i64 p)
 	return (sb.buf_high + (p - sb.offset_high));
 }
 
-static inline void put_u8(void *ss, int stream, uchar b)
+/* All put_u8/u32/vchars go to stream 0 */
+static inline void put_u8(void *ss, uchar b)
 {
-	if (unlikely(write_stream(ss, stream, &b, 1)))
+	if (unlikely(write_stream(ss, 0, &b, 1)))
 		fatal("Failed to put_u8\n");
 }
 
-static inline void put_u32(void *ss, int stream, uint32_t s)
+static inline void put_u32(void *ss, uint32_t s)
 {
-	if (unlikely(write_stream(ss, stream, (uchar *)&s, 4)))
+	if (unlikely(write_stream(ss, 0, (uchar *)&s, 4)))
 		fatal("Failed to put_u32\n");
 }
 
 /* Put a variable length of bytes dependant on how big the chunk is */
-static inline void put_vchars(void *ss, int stream, i64 s, int length)
+static inline void put_vchars(void *ss, i64 s, int length)
 {
 	int bytes;
 
@@ -180,14 +188,14 @@ static inline void put_vchars(void *ss, int stream, i64 s, int length)
 		int bits = bytes * 8;
 		uchar sb = (s >> bits) & (i64)0XFF;
 
-		put_u8(ss, stream, sb);
+		put_u8(ss, sb);
 	}
 }
 
 static void put_header(void *ss, uchar head, i64 len)
 {
-	put_u8(ss, 0, head);
-	put_vchars(ss, 0, len, 2);
+	put_u8(ss, head);
+	put_vchars(ss, len, 2);
 }
 
 static void put_match(struct rzip_state *st, i64 p, i64 offset, i64 len)
@@ -200,7 +208,7 @@ static void put_match(struct rzip_state *st, i64 p, i64 offset, i64 len)
 
 		ofs = (p - offset);
 		put_header(st->ss, 1, n);
-		put_vchars(st->ss, 0, ofs, st->chunk_bytes);
+		put_vchars(st->ss, ofs, st->chunk_bytes);
 
 		st->stats.matches++;
 		st->stats.match_bytes += n;
@@ -592,11 +600,6 @@ static void hash_search(struct rzip_state *st, double pct_base, double pct_multi
 		}
 	}
 
-	/* Fake that we got to 100% since we're done :D */
-	if (!STDIN)
-		print_progress("Total: 100%%  ");
-	print_progress("Chunk: 100%%\n");
-
 	if (MAX_VERBOSE)
 		show_distrib(st);
 
@@ -612,7 +615,7 @@ static void hash_search(struct rzip_state *st, double pct_base, double pct_multi
 	}
 
 	put_literal(st, 0, 0);
-	put_u32(st->ss, 0, st->cksum);
+	put_u32(st->ss, st->cksum);
 }
 
 
@@ -660,7 +663,7 @@ static void mmap_stdin(uchar *buf, struct rzip_state *st)
 			buf = mremap(buf, st->chunk_size, total, 0);
 			if (unlikely(buf == MAP_FAILED))
 				fatal("Failed to remap to smaller buf in mmap_stdin\n");
-			st->chunk_size = total;
+			st->mmap_size = st->chunk_size = total;
 			st->stdin_eof = 1;
 			break;
 		}
@@ -673,14 +676,18 @@ static void mmap_stdin(uchar *buf, struct rzip_state *st)
 static void init_sliding_mmap(struct rzip_state *st, int fd_in, i64 offset)
 {
 	/* Initialise the high buffer */
-	if (UNLIMITED) {
+	if (!STDIN) {
 		sb.high_length = 65536;
+		/* Round up to the next biggest page size */
+		if (sb.high_length % control.page_size)
+			sb.high_length += control.page_size - (sb.high_length % control.page_size);
 		sb.buf_high = (uchar *)mmap(NULL, sb.high_length, PROT_READ, MAP_SHARED, fd_in, offset);
 		if (unlikely(sb.buf_high == MAP_FAILED))
 			fatal("Unable to mmap buf_high in init_sliding_mmap\n");
 		sb.size_high = sb.high_length;
 		sb.offset_high = 0;
 	}
+	sb.low_top = 0;
 	sb.offset_low = 0;
 	sb.offset_search = 0;
 	sb.size_low = st->mmap_size;
@@ -695,17 +702,17 @@ static void rzip_chunk(struct rzip_state *st, int fd_in, int fd_out, i64 offset,
 {
 	init_sliding_mmap(st, fd_in, offset);
 
-	st->ss = open_stream_out(fd_out, NUM_STREAMS, st->chunk_size);
+	st->ss = open_stream_out(fd_out, NUM_STREAMS, st->chunk_size, st->chunk_bytes);
 	if (unlikely(!st->ss))
 		fatal("Failed to open streams in rzip_chunk\n");
 
-	print_verbose("Performing rzip pre-processing phase\n");
+	print_verbose("Beginning rzip pre-processing phase\n");
 	hash_search(st, pct_base, pct_multiple);
 
 	/* unmap buffer before closing and reallocating streams */
 	if (unlikely(munmap(sb.buf_low, sb.size_low)))
 		fatal("Failed to munmap in rzip_chunk\n");
-	if (UNLIMITED) {
+	if (!STDIN) {
 		if (unlikely(munmap(sb.buf_high, sb.size_high)))
 			fatal("Failed to munmap in rzip_chunk\n");
 	}
@@ -729,7 +736,7 @@ void rzip_fd(int fd_in, int fd_out)
 	struct timeval current, start, last;
 	struct stat s, s2;
 	struct rzip_state *st;
-	i64 len = 0, chunk_window, last_chunk = 0;
+	i64 len = 0, last_chunk = 0;
 	int pass = 0, passes;
 	unsigned int eta_hours, eta_minutes, eta_seconds, elapsed_hours,
 		     elapsed_minutes, elapsed_seconds;
@@ -753,17 +760,36 @@ void rzip_fd(int fd_in, int fd_out)
 	} else
 		control.st_size = 0;
 
+	/* Optimal use of ram involves using no more than 2/3 of it, so we
+	 * allocate 1/3 of it to the main buffer and use a sliding mmap
+	 * buffer to work on 2/3 ram size, leaving enough ram for the
+	 * compression backends */
+	control.max_mmap = control.ramsize / 3;
+
+	/* On 32 bits we can have a big window with sliding mmap, but can
+	 * not enable much per mmap/malloc */
+	if (BITS32)
+		control.max_mmap = MIN(control.max_mmap, two_gig);
+	round_to_page(&control.max_mmap);
+
+	/* Set maximum chunk size to proportion of ram according to mode */
+	if (MAXRAM && !STDIN)
+		control.max_chunk = control.ramsize / 2 * 3;
+	else
+		control.max_chunk = control.ramsize / 3 * 2;
+	if (UNLIMITED)
+		control.max_chunk = control.st_size;
 	if (control.window)
-		chunk_window = control.window * CHUNK_MULTIPLE;
-	else {
-		if (STDIN)
-			chunk_window = control.ramsize;
-		else
-			chunk_window = len;
-	}
-	if (chunk_window < len)
-		chunk_window -= chunk_window % control.page_size;
-	st->chunk_size = chunk_window;
+		control.max_chunk = MIN(control.max_chunk, control.window * CHUNK_MULTIPLE);
+	round_to_page(&control.max_chunk);
+	control.max_mmap = MIN(control.max_mmap, control.max_chunk);
+
+	if (!STDIN)
+		st->chunk_size = MIN(control.max_chunk, len);
+	else
+		st->chunk_size = control.max_mmap;
+	if (st->chunk_size < len)
+		round_to_page(&st->chunk_size);
 
 	st->level = &levels[control.compression_level];
 	st->fd_in = fd_in;
@@ -772,85 +798,70 @@ void rzip_fd(int fd_in, int fd_out)
 
 	init_hash_indexes(st);
 
-	passes = 1 + s.st_size / chunk_window;
+	passes = 0;
 
 	/* set timers and chunk counter */
 	last.tv_sec = last.tv_usec = 0;
 	gettimeofday(&start, NULL);
 
+	prepare_streamout_threads();
+
 	while (len > 0 || (STDIN && !st->stdin_eof)) {
 		double pct_base, pct_multiple;
 		i64 offset = s.st_size - len;
 		int bits = 8;
 
-		/* Flushing the dirty data will decrease our chances of
-		 * running out of memory when we allocate ram again on the
-		 * next chunk. It will also prevent thrashing on-disk due to
-		 * concurrent reads and writes if we're on the same device. */
-		if (last_chunk)
-			print_verbose("Flushing data to disk.\n");
-		fsync(fd_out);
-
-		if (st->chunk_size > len && !STDIN)
-			st->chunk_size = len;
-		st->mmap_size = st->chunk_size;
-		if (BITS32 && st->mmap_size > two_gig) {
-			print_verbose("Limiting to 2GB due to 32 bit limitations\n");
-			st->mmap_size = two_gig;
+		st->chunk_size = control.max_chunk;
+		st->mmap_size = control.max_mmap;
+		if (!STDIN) {
+			st->chunk_size = MIN(st->chunk_size, len);
+			st->mmap_size = MIN(st->mmap_size, len);
 		}
 
 retry:
-		/* Mmapping anonymously first will tell us how much ram we can use in
-		 * advance and zeroes it which has a defragmenting effect on ram
-		 * before the real read in. */
-		sb.buf_low = mmap(NULL, st->mmap_size, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
-		/* Better to shrink the window to the largest size that works than fail */
-		if (sb.buf_low == MAP_FAILED) {
-			st->mmap_size = st->mmap_size / 10 * 9;
-			st->mmap_size -= st->mmap_size % control.page_size;
-			if (unlikely(!st->mmap_size))
-				fatal("Unable to mmap any ram\n");
-			goto retry;
-		}
-
-		/* NOTE the buf is saved here for STDIN mode */
-		if (!STDIN) {
-			if (unlikely(munmap(sb.buf_low, st->mmap_size)))
-				fatal("Failed to munmap\n");
-		}
-
-		if (!MAXRAM) {
-			print_maxverbose("Succeeded in allocating %lld sized mmap\n", st->mmap_size);
-			if (!UNLIMITED)
-				st->chunk_size = st->mmap_size;
-		} else
-			st->mmap_size = st->chunk_size;
-
-		if (!STDIN) {
-			/* The buf is saved here for !STDIN mode */
-			sb.buf_low = (uchar *)mmap(sb.buf_low, st->mmap_size, PROT_READ, MAP_SHARED, fd_in, offset);
+		if (STDIN) {
+			/* NOTE the buf is saved here for STDIN mode */
+			sb.buf_low = mmap(NULL, st->mmap_size, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
+			/* Better to shrink the window to the largest size that works than fail */
 			if (sb.buf_low == MAP_FAILED) {
-				if (unlikely(!MAXRAM))
-					fatal("Failed to remap ram\n");
 				st->mmap_size = st->mmap_size / 10 * 9;
-				st->mmap_size -= st->mmap_size % control.page_size;
+				round_to_page(&st->mmap_size);
 				if (unlikely(!st->mmap_size))
 					fatal("Unable to mmap any ram\n");
 				goto retry;
 			}
-		} else
+			st->chunk_size = st->mmap_size;
 			mmap_stdin(sb.buf_low, st);
+		} else {
+			/* NOTE The buf is saved here for !STDIN mode */
+			sb.buf_low = (uchar *)mmap(sb.buf_low, st->mmap_size, PROT_READ, MAP_SHARED, fd_in, offset);
+			if (sb.buf_low == MAP_FAILED) {
+				st->mmap_size = st->mmap_size / 10 * 9;
+				round_to_page(&st->mmap_size);
+				if (unlikely(!st->mmap_size))
+					fatal("Unable to mmap any ram\n");
+				goto retry;
+			}
+			if (st->mmap_size < st->chunk_size)
+				print_maxverbose("Enabling sliding mmap mode and using mmap of %lld bytes with window of %lld bytes\n", st->mmap_size, st->chunk_size);
+		}
+		print_maxverbose("Succeeded in testing %lld sized mmap for rzip pre-processing\n", st->mmap_size);
 
-		if (MAXRAM)
-			print_maxverbose("Succeeded in allocating %lld sized mmap\n", st->mmap_size);
+		if (st->chunk_size > control.ramsize)
+			print_verbose("Compression window is larger than ram, will proceed with unlimited mode possibly much slower\n");
 
-		if (st->mmap_size < st->chunk_size)
-			print_verbose("Compression window is larger than ram allocated, will proceed with unlimited mode possibly much slower\n");
+		if (!passes && !STDIN) {
+			passes = s.st_size / st->chunk_size + !!(s.st_size % st->chunk_size);
+			if (passes == 1)
+				print_verbose("Will take 1 pass\n");
+			else
+				print_verbose("Will take %d passes\n", passes);
+		}
 
 		sb.orig_offset = offset;
 		print_maxverbose("Chunk size: %lld\n", st->chunk_size);
 
-		/* Determine the chunk byte width and write it to the file
+		/* Determine the chunk byte width to write to the file
 		 * This allows archives of different chunk sizes to have
 		 * optimal byte width entries. When working with stdin we
 		 * won't know in advance how big it is so it will always be
@@ -861,8 +872,6 @@ retry:
 		if (bits % 8)
 			st->chunk_bytes++;
 		print_maxverbose("Byte width: %d\n", st->chunk_bytes);
-		if (unlikely(write(fd_out, &st->chunk_bytes, 1) != 1))
-			fatal("Failed to write chunk_bytes size in rzip_fd\n");
 
 		pct_base = (100.0 * (s.st_size - len)) / s.st_size;
 		pct_multiple = ((double)st->chunk_size) / s.st_size;
@@ -885,18 +894,22 @@ retry:
 					pass, passes, elapsed_hours, elapsed_minutes, elapsed_seconds,
 					eta_hours, eta_minutes, eta_seconds, chunkmbs);
 			else
-				print_verbose("\nPass %d / %d -- Elapsed Time: %02d:%02d:%02d. Compress Speed: %3.3fMB/s.\n",
-					pass, passes, elapsed_hours, elapsed_minutes, elapsed_seconds, chunkmbs);
+				print_verbose("\nPass %d -- Elapsed Time: %02d:%02d:%02d. Compress Speed: %3.3fMB/s.\n",
+					pass, elapsed_hours, elapsed_minutes, elapsed_seconds, chunkmbs);
 		}
 		last.tv_sec = current.tv_sec;
 		last.tv_usec = current.tv_usec;
 		rzip_chunk(st, fd_in, fd_out, offset, pct_base, pct_multiple);
-		/* st->chunk_bytes may be shrunk in rzip_chunk */
+		/* st->chunk_size may be shrunk in rzip_chunk */
 		last_chunk = st->chunk_size;
 		len -= st->chunk_size;
 	}
 
+	close_streamout_threads();
+
 	gettimeofday(&current, NULL);
+	if (STDIN)
+		s.st_size = control.st_size;
 	chunkmbs = (s.st_size / 1024 / 1024) / ((double)(current.tv_sec-start.tv_sec)? : 1);
 
 	fstat(fd_out, &s2);
