@@ -35,6 +35,17 @@ struct compress_thread{
 	int stream;
 } *cthread;
 
+struct uncomp_thread{
+	uchar *s_buf;
+	i64 u_len, c_len;
+	i64 last_head;
+	uchar c_type;
+	sem_t complete;
+	sem_t ready;	/* Taken this thread's data so it can die */
+	sem_t free;
+	int stream;
+} *ucthread;
+
 void init_sem(sem_t *sem)
 {
 	if (unlikely(sem_init(sem, 0, 0)))
@@ -61,6 +72,20 @@ retry:
 	}
 }
 
+static inline int trywait_sem(sem_t *s)
+{
+	int ret;
+
+retry:
+	if ((ret = sem_trywait(s)) == -1) {
+		if (errno == EINTR)
+			goto retry;
+		if (errno != EAGAIN)
+			fatal("sem_trywait");
+	}
+	return ret;
+}
+
 static inline void destroy_sem(sem_t *s)
 {
 	if (unlikely(sem_destroy(s)))
@@ -140,24 +165,29 @@ static inline int fake_open_memstream_update_buffer(FILE *fp, uchar **buf, size_
   length in c_len
 */
 
-static void zpaq_compress_buf(struct compress_thread *cthread, long thread)
+static int zpaq_compress_buf(struct compress_thread *cthread, long thread)
 {
 	uchar *c_buf = NULL;
 	size_t dlen = 0;
 	FILE *in, *out;
 
 	if (!lzo_compresses(cthread->s_buf, cthread->s_len))
-		return;
+		return 0;
 
 	in = fmemopen(cthread->s_buf, cthread->s_len, "r");
-	if (unlikely(!in))
-		fatal("Failed to fmemopen in zpaq_compress_buf\n");
+	if (unlikely(!in)) {
+		print_maxverbose("Failed to fmemopen in zpaq_compress_buf\n");
+		return -1;
+	}
 	out = open_memstream((char **)&c_buf, &dlen);
-	if (unlikely(!out))
-		fatal("Failed to open_memstream in zpaq_compress_buf\n");
+	if (unlikely(!out)) {
+		fclose(in);
+		print_maxverbose("Failed to open_memstream in zpaq_compress_buf\n");
+		return -1;
+	}
 
-	zpipe_compress(in, out, control.msgout, cthread->s_len, (int)(SHOW_PROGRESS),
-		       thread);
+	zpipe_compress(in, out, control.msgout, cthread->s_len,
+		       (int)(SHOW_PROGRESS), thread);
 
 	if (unlikely(memstream_update_buffer(out, &c_buf, &dlen)))
 	        fatal("Failed to memstream_update_buffer in zpaq_compress_buf");
@@ -169,152 +199,183 @@ static void zpaq_compress_buf(struct compress_thread *cthread, long thread)
 		print_maxverbose("Incompressible block\n");
 		/* Incompressible, leave as CTYPE_NONE */
 		free(c_buf);
-		return;
+		return 0;
 	}
 
 	cthread->c_len = dlen;
 	free(cthread->s_buf);
 	cthread->s_buf = c_buf;
 	cthread->c_type = CTYPE_ZPAQ;
+	return 0;
 }
 
-static void bzip2_compress_buf(struct compress_thread *cthread)
+static int bzip2_compress_buf(struct compress_thread *cthread)
 {
 	u32 dlen = cthread->s_len;
 	uchar *c_buf;
 
 	if (!lzo_compresses(cthread->s_buf, cthread->s_len))
-		return;
+		return 0;
 
 	c_buf = malloc(dlen);
-	if (!c_buf)
-		return;
+	if (!c_buf) {
+		print_maxverbose("Unable to allocate c_buf\n");
+		return -1;
+	}
 
 	if (BZ2_bzBuffToBuffCompress((char *)c_buf, &dlen,
 		(char *)cthread->s_buf, cthread->s_len,
 		control.compression_level, 0, control.compression_level * 10) != BZ_OK) {
 			free(c_buf);
-			return;
+			print_maxverbose("BZ2 compress failed\n");
+			return -1;
 	}
 
-	if (dlen >= cthread->c_len) {
+	if (unlikely(dlen >= cthread->c_len)) {
 		print_maxverbose("Incompressible block\n");
 		/* Incompressible, leave as CTYPE_NONE */
 		free(c_buf);
-		return;
+		return 0;
 	}
 
 	cthread->c_len = dlen;
 	free(cthread->s_buf);
 	cthread->s_buf = c_buf;
 	cthread->c_type = CTYPE_BZIP2;
+	return 0;
 }
 
-static void gzip_compress_buf(struct compress_thread *cthread)
+static int gzip_compress_buf(struct compress_thread *cthread)
 {
 	unsigned long dlen = cthread->s_len;
 	uchar *c_buf;
 
 	c_buf = malloc(dlen);
-	if (!c_buf)
-		return;
+	if (!c_buf) {
+		print_maxverbose("Unable to allocate c_buf\n");
+		return -1;
+	}
 
 	if (compress2(c_buf, &dlen, cthread->s_buf, cthread->s_len,
 		control.compression_level) != Z_OK) {
 			free(c_buf);
-			return;
+			print_maxverbose("compress2 failed\n");
+			return -1;
 	}
 
 	if ((i64)dlen >= cthread->c_len) {
 		print_maxverbose("Incompressible block\n");
 		/* Incompressible, leave as CTYPE_NONE */
 		free(c_buf);
-		return;
+		return 0;
 	}
 
 	cthread->c_len = dlen;
 	free(cthread->s_buf);
 	cthread->s_buf = c_buf;
 	cthread->c_type = CTYPE_GZIP;
+	return 0;
 }
 
-static void lzma_compress_buf(struct compress_thread *cthread)
+static int lzma_compress_buf(struct compress_thread *cthread)
 {
+	int lzma_level, lzma_ret;
 	size_t prop_size = 5; /* return value for lzma_properties */
 	uchar *c_buf;
 	size_t dlen;
-	int lzma_ret;
 
 	if (!lzo_compresses(cthread->s_buf, cthread->s_len))
-		return;
+		return 0;
 
+	/* only 7 levels with lzma, scale them */
+	lzma_level = control.compression_level * 7 / 9 ? : 1;
+
+	print_verbose("Starting lzma back end compression thread...\n");
+retry:
 	dlen = cthread->s_len;
 	c_buf = malloc(dlen);
 	if (!c_buf)
-		return;
+		return -1;
 
-	print_progress("Starting lzma back end compression thread...\n");
 	/* with LZMA SDK 4.63, we pass compression level and threads only
 	 * and receive properties in control->lzma_properties */
 
 	lzma_ret = LzmaCompress(c_buf, &dlen, cthread->s_buf,
-		(size_t)cthread->s_len, control.lzma_properties, &prop_size, control.compression_level,
-			      0, /* dict size. set default */
-			      -1, -1, -1, -1, /* lc, lp, pb, fb */
-			      control.threads);
+		(size_t)cthread->s_len, control.lzma_properties, &prop_size,
+				lzma_level,
+				0, /* dict size. set default, choose by level */
+				-1, -1, -1, -1, /* lc, lp, pb, fb */
+				control.threads);
 	if (lzma_ret != SZ_OK) {
 		switch (lzma_ret) {
 			case SZ_ERROR_MEM:
-				print_err("\nLZMA ERROR: %d. Try a smaller compression window.\n", SZ_ERROR_MEM);
 				break;
 			case SZ_ERROR_PARAM:
-				print_err("\nLZMA Parameter ERROR: %d. This should not happen.\n", SZ_ERROR_PARAM);
+				print_err("LZMA Parameter ERROR: %d. This should not happen.\n", SZ_ERROR_PARAM);
 				break;
 			case SZ_ERROR_OUTPUT_EOF:
-				print_maxverbose("\nHarmless LZMA Output Buffer Overflow error: %d. Incompressible block.\n", SZ_ERROR_OUTPUT_EOF);
+				print_maxverbose("Harmless LZMA Output Buffer Overflow error: %d. Incompressible block.\n", SZ_ERROR_OUTPUT_EOF);
 				break;
 			case SZ_ERROR_THREAD:
-				print_err("\nLZMA Multi Thread ERROR: %d. This should not happen.\n", SZ_ERROR_THREAD);
+				print_err("LZMA Multi Thread ERROR: %d. This should not happen.\n", SZ_ERROR_THREAD);
 				break;
 			default:
 				print_err("Unidentified LZMA ERROR: %d. This should not happen.\n", lzma_ret);
 				break;
 		}
 		/* can pass -1 if not compressible! Thanks Lasse Collin */
-		print_maxverbose("Incompressible block\n");
 		free(c_buf);
-		return;
+		if (lzma_ret == SZ_ERROR_MEM) {
+			if (lzma_level > 1) {
+				lzma_level--;
+				print_verbose("LZMA Warning: %d. Can't allocate enough RAM for compression window, trying smaller.\n", SZ_ERROR_MEM);
+				goto retry;
+			}
+			/* lzma compress can be fragile on 32 bit. If it fails,
+			 * fall back to bzip2 compression so the block doesn't
+			 * remain uncompressed */
+			print_verbose("Unable to allocate enough RAM for any sized compression window, falling back to bzip2 compression.\n");
+			return bzip2_compress_buf(cthread);
+		} else if (lzma_ret == SZ_ERROR_OUTPUT_EOF)
+			return 0;
+		return -1;
 	}
-	if ((i64)dlen >= cthread->c_len) {
+
+	if (unlikely((i64)dlen >= cthread->c_len)) {
 		/* Incompressible, leave as CTYPE_NONE */
 		print_maxverbose("Incompressible block\n");
 		free(c_buf);
-		return;
+		return 0;
 	}
 
 	cthread->c_len = dlen;
 	free(cthread->s_buf);
 	cthread->s_buf = c_buf;
 	cthread->c_type = CTYPE_LZMA;
+	return 0;
 }
 
-static void lzo_compress_buf(struct compress_thread *cthread)
+static int lzo_compress_buf(struct compress_thread *cthread)
 {
 	lzo_uint in_len = cthread->s_len;
 	lzo_uint dlen = in_len + in_len / 16 + 64 + 3;
 	lzo_int return_var;	/* lzo1x_1_compress does not return anything but LZO_OK */
 	lzo_bytep wrkmem;
 	uchar *c_buf;
+	int ret = -1;
 
 	wrkmem = (lzo_bytep) malloc(LZO1X_1_MEM_COMPRESS);
-	if (wrkmem == NULL)
-		return;
+	if (unlikely(wrkmem == NULL)) {
+		print_maxverbose("Failed to malloc wkmem\n");
+		return ret;
+	}
 
 	c_buf = malloc(dlen);
 	if (!c_buf)
 		goto out_free;
 
 	return_var = lzo1x_1_compress(cthread->s_buf, in_len, c_buf, &dlen, wrkmem);
+	ret = 0;
 
 	if (dlen >= in_len){
 		/* Incompressible, leave as CTYPE_NONE */
@@ -329,6 +390,7 @@ static void lzo_compress_buf(struct compress_thread *cthread)
 	cthread->c_type = CTYPE_LZO;
 out_free:
 	free(wrkmem);
+	return ret;
 }
 
 /*
@@ -339,13 +401,13 @@ out_free:
   try to decompress a buffer. Return 0 on success and -1 on failure.
 */
 
-static int zpaq_decompress_buf(struct stream *s)
+static int zpaq_decompress_buf(struct uncomp_thread *ucthread, long thread)
 {
 	uchar *c_buf = NULL;
 	size_t dlen = 0;
 	FILE *in, *out;
 
-	in = fmemopen(s->buf, s->buflen, "r");
+	in = fmemopen(ucthread->s_buf, ucthread->u_len, "r");
 	if (unlikely(!in)) {
 		print_err("Failed to fmemopen in zpaq_decompress_buf\n");
 		return -1;
@@ -356,136 +418,161 @@ static int zpaq_decompress_buf(struct stream *s)
 		return -1;
 	}
 
-	zpipe_decompress(in, out, control.msgout, s->buflen, (int)(SHOW_PROGRESS));
+	zpipe_decompress(in, out, control.msgout, ucthread->u_len, (int)(SHOW_PROGRESS), thread);
 
 	if (unlikely(memstream_update_buffer(out, &c_buf, &dlen)))
 	        fatal("Failed to memstream_update_buffer in zpaq_decompress_buf");
 
 	fclose(in);
 	fclose(out);
-	free(s->buf);
-	s->buf = c_buf;
 
-	if (unlikely((i64)dlen != s->buflen)) {
-		print_err("Inconsistent length after decompression. Got %lld bytes, expected %lld\n", (i64)dlen, s->buflen);
+	if (unlikely((i64)dlen != ucthread->u_len)) {
+		print_err("Inconsistent length after decompression. Got %lld bytes, expected %lld\n", (i64)dlen, ucthread->u_len);
 		return -1;
 	}
 
+	free(ucthread->s_buf);
+	ucthread->s_buf = c_buf;
+
 	return 0;
 }
 
-static int bzip2_decompress_buf(struct stream *s, i64 c_len)
+static int bzip2_decompress_buf(struct uncomp_thread *ucthread)
 {
-	u32 dlen = s->buflen;
+	u32 dlen = ucthread->u_len;
+	int ret = 0, bzerr;
 	uchar *c_buf;
-	int bzerr;
 
-	c_buf = s->buf;
-	s->buf = malloc(dlen);
-	if (unlikely(!s->buf)) {
+	c_buf = ucthread->s_buf;
+	ucthread->s_buf = malloc(dlen);
+	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %d bytes for decompression\n", dlen);
-		return -1;
+		ret = -1;
+		goto out;
 	}
 
-	bzerr = BZ2_bzBuffToBuffDecompress((char*)s->buf, &dlen, (char*)c_buf, c_len, 0, 0);
+	bzerr = BZ2_bzBuffToBuffDecompress((char*)ucthread->s_buf, &dlen, (char*)c_buf, ucthread->c_len, 0, 0);
 	if (unlikely(bzerr != BZ_OK)) {
 		print_err("Failed to decompress buffer - bzerr=%d\n", bzerr);
-		return -1;
+		ret = -1;
+		goto out_free;
 	}
 
-	if (unlikely(dlen != s->buflen)) {
-		print_err("Inconsistent length after decompression. Got %d bytes, expected %lld\n", dlen, s->buflen);
-		return -1;
+	if (unlikely(dlen != ucthread->u_len)) {
+		print_err("Inconsistent length after decompression. Got %d bytes, expected %lld\n", dlen, ucthread->u_len);
+		ret = -1;
 	}
 
+out_free:
 	free(c_buf);
-	return 0;
+out:
+	if (ret == -1)
+		ucthread->s_buf = c_buf;
+	return ret;
 }
 
-static int gzip_decompress_buf(struct stream *s, i64 c_len)
+static int gzip_decompress_buf(struct uncomp_thread *ucthread)
 {
-	unsigned long dlen = s->buflen;
+	unsigned long dlen = ucthread->u_len;
+	int ret = 0, gzerr;
 	uchar *c_buf;
-	int gzerr;
 
-	c_buf = s->buf;
-	s->buf = malloc(dlen);
-	if (unlikely(!s->buf)) {
+	c_buf = ucthread->s_buf;
+	ucthread->s_buf = malloc(dlen);
+	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %ld bytes for decompression\n", dlen);
-		return -1;
+		ret = -1;
+		goto out;
 	}
 
-	gzerr = uncompress(s->buf, &dlen, c_buf, c_len);
+	gzerr = uncompress(ucthread->s_buf, &dlen, c_buf, ucthread->c_len);
 	if (unlikely(gzerr != Z_OK)) {
 		print_err("Failed to decompress buffer - bzerr=%d\n", gzerr);
-		return -1;
+		ret = -1;
+		goto out_free;
 	}
 
-	if (unlikely((i64)dlen != s->buflen)) {
-		print_err("Inconsistent length after decompression. Got %ld bytes, expected %lld\n", dlen, s->buflen);
-		return -1;
+	if (unlikely((i64)dlen != ucthread->u_len)) {
+		print_err("Inconsistent length after decompression. Got %ld bytes, expected %lld\n", dlen, ucthread->u_len);
+		ret = -1;
 	}
 
+out_free:
 	free(c_buf);
-	return 0;
+out:
+	if (ret == -1)
+		ucthread->s_buf = c_buf;
+	return ret;
 }
 
-static int lzma_decompress_buf(struct stream *s, size_t c_len)
+static int lzma_decompress_buf(struct uncomp_thread *ucthread)
 {
-	size_t dlen = (size_t)s->buflen;
+	size_t dlen = (size_t)ucthread->u_len;
+	int ret = 0, lzmaerr;
 	uchar *c_buf;
-	int lzmaerr;
 
-	c_buf = s->buf;
-	s->buf = malloc(dlen);
-	if (unlikely(!s->buf)) {
+	c_buf = ucthread->s_buf;
+	ucthread->s_buf = malloc(dlen);
+	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %lldd bytes for decompression\n", (i64)dlen);
-		return -1;
+		ret = -1;
+		goto out;
 	}
 
 	/* With LZMA SDK 4.63 we pass control.lzma_properties
 	 * which is needed for proper uncompress */
-	lzmaerr = LzmaUncompress(s->buf, &dlen, c_buf, &c_len, control.lzma_properties, 5);
+	lzmaerr = LzmaUncompress(ucthread->s_buf, &dlen, c_buf, (SizeT *)&ucthread->c_len, control.lzma_properties, 5);
 	if (unlikely(lzmaerr)) {
 		print_err("Failed to decompress buffer - lzmaerr=%d\n", lzmaerr);
-		return -1;
+		ret = -1;
+		goto out_free;
 	}
 
-	if (unlikely((i64)dlen != s->buflen)) {
-		print_err("Inconsistent length after decompression. Got %lld bytes, expected %lld\n", (i64)dlen, s->buflen);
-		return -1;
+	if (unlikely((i64)dlen != ucthread->u_len)) {
+		print_err("Inconsistent length after decompression. Got %lld bytes, expected %lld\n", (i64)dlen, ucthread->u_len);
+		ret = -1;
 	}
 
+out_free:
 	free(c_buf);
-	return 0;
+out:
+	if (ret == -1)
+		ucthread->s_buf = c_buf;
+	return ret;
 }
 
-static int lzo_decompress_buf(struct stream *s, i64 c_len)
+static int lzo_decompress_buf(struct uncomp_thread *ucthread)
 {
-	lzo_uint dlen = s->buflen;
+	lzo_uint dlen = ucthread->u_len;
+	int ret = 0, lzerr;
 	uchar *c_buf;
-	int lzerr;
 
-	c_buf = s->buf;
-	s->buf = malloc(dlen);
-	if (unlikely(!s->buf)) {
+	c_buf = ucthread->s_buf;
+	ucthread->s_buf = malloc(dlen);
+	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %lu bytes for decompression\n", (unsigned long)dlen);
-		return -1;
+		ret = -1;
+		goto out;
 	}
 
-	lzerr = lzo1x_decompress((uchar*)c_buf,c_len,(uchar*)s->buf,&dlen,NULL);
+	lzerr = lzo1x_decompress((uchar*)c_buf, ucthread->c_len, (uchar*)ucthread->s_buf, &dlen,NULL);
 	if (unlikely(lzerr != LZO_E_OK)) {
 		print_err("Failed to decompress buffer - lzerr=%d\n", lzerr);
-		return -1;
+		ret = -1;
+		goto out_free;
 	}
 
-	if (unlikely((i64)dlen != s->buflen)) {
-		print_err("Inconsistent length after decompression. Got %lu bytes, expected %lld\n", (unsigned long)dlen, s->buflen);
-		return -1;
+	if (unlikely((i64)dlen != ucthread->u_len)) {
+		print_err("Inconsistent length after decompression. Got %lu bytes, expected %lld\n", (unsigned long)dlen, ucthread->u_len);
+		ret = -1;
 	}
 
+out_free:
 	free(c_buf);
-	return 0;
+out:
+	if (ret == -1)
+		ucthread->s_buf = c_buf;
+	return ret;
 }
 
 /* WORK FUNCTIONS */
@@ -619,25 +706,19 @@ static int seekto(struct stream_info *sinfo, i64 pos)
 	return 0;
 }
 
-/* open a set of output streams, compressing with the given
-   compression level and algorithm */
-void *open_stream_out(int f, int n, i64 limit)
+static pthread_t *threads;
+
+void prepare_streamout_threads(void)
 {
-	unsigned cwindow = control.window;
-	struct stream_info *sinfo;
-	uchar *testmalloc;
 	int i;
 
-	sinfo = malloc(sizeof(*sinfo));
-	if (unlikely(!sinfo))
-		return NULL;
-
-	sinfo->bufsize = limit;
-	sinfo->thread_no = 0;
+	threads = calloc(sizeof(pthread_t), control.threads);
+	if (unlikely(!threads))
+		fatal("Unable to calloc threads in prepare_streamout_threads\n");
 
 	cthread = calloc(sizeof(struct compress_thread), control.threads);
 	if (unlikely(!cthread))
-		fatal("Unable to calloc cthread in open_stream_out\n");
+		fatal("Unable to calloc cthread in prepare_streamout_threads\n");
 
 	for (i = 0; i < control.threads; i++) {
 		init_sem(&cthread[i].complete);
@@ -654,26 +735,45 @@ void *open_stream_out(int f, int n, i64 limit)
 	/* Signal thread 0 that it can start */
 	if (control.threads > 1)
 		post_sem(&cthread[control.threads - 1].complete);
+}
 
+void close_streamout_threads(void)
+{
+	int i;
+
+	for (i = 0; i < control.threads; i++) {
+		wait_sem(&cthread[i].free);
+		destroy_sem(&cthread[i].complete);
+		destroy_sem(&cthread[i].free);
+	}
+	free(cthread);
+	free(threads);
+}
+
+/* open a set of output streams, compressing with the given
+   compression level and algorithm */
+void *open_stream_out(int f, int n, i64 limit, char cbytes)
+{
+	struct stream_info *sinfo;
+	uchar *testmalloc;
+	i64 testsize;
+	int i;
+
+	sinfo = calloc(sizeof(struct stream_info), 1);
+	if (unlikely(!sinfo))
+		return NULL;
+
+	sinfo->bufsize = limit;
+
+	sinfo->chunk_bytes = cbytes;
 	sinfo->num_streams = n;
-	sinfo->cur_pos = 0;
 	sinfo->fd = f;
 
-	if (BITS32) {
-		/* Largest window supported on 32bit is 600MB */
-		if (!cwindow || cwindow > 6)
-			cwindow = 6;
-	}
+	/* Serious limits imposed on 32 bit capabilities */
+	if (BITS32)
+		limit = MIN(limit, two_gig / 3);
 
-	/* No point making the stream larger than the amount of data */
-	if (cwindow)
-		sinfo->bufsize = MIN(STREAM_BUFSIZE * 10 * cwindow, limit);
-	else
-		sinfo->bufsize = limit;
-
-	sinfo->initial_pos = lseek(f, 0, SEEK_CUR);
-
-	sinfo->s = (struct stream *)calloc(sizeof(sinfo->s[0]), n);
+	sinfo->s = calloc(sizeof(struct stream), n);
 	if (unlikely(!sinfo->s)) {
 		free(sinfo);
 		return NULL;
@@ -683,12 +783,16 @@ void *open_stream_out(int f, int n, i64 limit)
 	 * ram. We need enough for the 2 streams and for the compression
 	 * backend at most, being conservative. */
 retest_malloc:
-	testmalloc = malloc(sinfo->bufsize * (n + 1));
+	testsize = limit * (n + 1);
+	testmalloc = malloc(testsize);
 	if (!testmalloc) {
-		sinfo->bufsize = sinfo->bufsize / 10 * 9;
+		limit = limit / 10 * 9;
 		goto retest_malloc;
 	}
 	free(testmalloc);
+	print_maxverbose("Succeeded in testing %lld sized malloc for back end compression\n", testsize);
+
+	sinfo->bufsize = limit;
 
 	/* Make the bufsize no smaller than STREAM_BUFSIZE. Round up the
 	 * bufsize to fit X threads into it */
@@ -708,15 +812,6 @@ retest_malloc:
 			fatal("Unable to malloc buffer of size %lld in open_stream_out\n", sinfo->bufsize);
 	}
 
-	/* write the initial headers */
-	for (i = 0; i < n; i++) {
-		sinfo->s[i].last_head = sinfo->cur_pos + 17;
-		write_u8(sinfo->fd, CTYPE_NONE);
-		write_i64(sinfo->fd, 0);
-		write_i64(sinfo->fd, 0);
-		write_i64(sinfo->fd, 0);
-		sinfo->cur_pos += 25;
-	}
 	return (void *)sinfo;
 }
 
@@ -724,27 +819,50 @@ retest_malloc:
 void *open_stream_in(int f, int n)
 {
 	struct stream_info *sinfo;
+	int total_threads, i;
 	i64 header_length;
-	int i;
 
-	sinfo = calloc(sizeof(*sinfo), 1);
+	sinfo = calloc(sizeof(struct stream_info), 1);
 	if (unlikely(!sinfo))
 		return NULL;
 
+	total_threads = control.threads + 1;
+	threads = calloc(sizeof(pthread_t), total_threads);
+	if (unlikely(!threads))
+		return NULL;
+
+	ucthread = calloc(sizeof(struct uncomp_thread), total_threads);
+	if (unlikely(!ucthread))
+		fatal("Unable to calloc cthread in open_stream_in\n");
+
+	for (i = 0; i < total_threads; i++) {
+		init_sem(&ucthread[i].complete);
+		init_sem(&ucthread[i].free);
+		post_sem(&ucthread[i].free);
+		init_sem(&ucthread[i].ready);
+	}
+
 	sinfo->num_streams = n;
 	sinfo->fd = f;
 	sinfo->initial_pos = lseek(f, 0, SEEK_CUR);
 
-	sinfo->s = (struct stream *)calloc(sizeof(sinfo->s[0]), n);
+	sinfo->s = calloc(sizeof(struct stream), n);
 	if (unlikely(!sinfo->s)) {
 		free(sinfo);
 		return NULL;
 	}
 
+	sinfo->s[0].total_threads = 1;
+	sinfo->s[1].total_threads = control.threads;
+
 	for (i = 0; i < n; i++) {
 		uchar c;
 		i64 v1, v2;
 
+		sinfo->s[i].base_thread = i;
+		sinfo->s[i].uthread_no = sinfo->s[i].base_thread;
+		sinfo->s[i].unext_thread = sinfo->s[i].base_thread;
+
 again:
 		if (unlikely(read_u8(f, &c)))
 			goto failed;
@@ -806,57 +924,94 @@ failed:
 
 /* Enter with s_buf allocated,s_buf points to the compressed data after the
  * backend compression and is then freed here */
-void *compthread(void *t)
+static void *compthread(void *t)
 {
 	long i = (long)t;
 	struct compress_thread *cti = &cthread[i];
+	struct stream_info *ctis = cti->sinfo;
+	int waited = 0, ret = 0;
+
+	if (unlikely(setpriority(PRIO_PROCESS, 0, control.nice_val) == -1))
+		print_err("Warning, unable to set nice value on thread\n");
 
 	cti->c_type = CTYPE_NONE;
 	cti->c_len = cti->s_len;
 
+retry:
 	if (!NO_COMPRESS && cti->c_len) {
 		if (LZMA_COMPRESS)
-			lzma_compress_buf(cti);
+			ret = lzma_compress_buf(cti);
 		else if (LZO_COMPRESS)
-			lzo_compress_buf(cti);
+			ret = lzo_compress_buf(cti);
 		else if (BZIP2_COMPRESS)
-			bzip2_compress_buf(cti);
+			ret = bzip2_compress_buf(cti);
 		else if (ZLIB_COMPRESS)
-			gzip_compress_buf(cti);
+			ret = gzip_compress_buf(cti);
 		else if (ZPAQ_COMPRESS)
-			zpaq_compress_buf(cti, i);
+			ret = zpaq_compress_buf(cti, i);
 		else fatal("Dunno wtf compression to use!\n");
 	}
 
-	if (control.threads > 1)
-		wait_sem(&cthread[cti->wait_on].complete);
+	/* If compression fails for whatever reason multithreaded, then wait
+	 * for the previous thread to finish, serialising the work to decrease
+	 * the memory requirements, increasing the chance of success */
+	if (ret){
+		if (waited)
+			fatal("Failed to compress in compthread\n");
+		print_maxverbose("Unable to compress in parallel, waiting for previous thread to complete before trying again\n");
+	}
 
-	if (unlikely(seekto(cti->sinfo, cti->sinfo->s[cti->stream].last_head)))
+	if (control.threads > 1 && !waited)
+		wait_sem(&cthread[cti->wait_on].complete);
+	waited = 1;
+	if (ret)
+		goto retry;
+
+	if (!ctis->chunks++) {
+		int j;
+
+		/* Write chunk bytes of this block */
+		write_u8(ctis->fd, ctis->chunk_bytes);
+
+		/* First chunk of this stream, write headers */
+		ctis->initial_pos = lseek(ctis->fd, 0, SEEK_CUR);
+
+		for (j = 0; j < ctis->num_streams; j++) {
+			ctis->s[j].last_head = ctis->cur_pos + 17;
+			write_u8(ctis->fd, CTYPE_NONE);
+			write_i64(ctis->fd, 0);
+			write_i64(ctis->fd, 0);
+			write_i64(ctis->fd, 0);
+			ctis->cur_pos += 25;
+		}
+	}
+
+	if (unlikely(seekto(ctis, ctis->s[cti->stream].last_head)))
 		fatal("Failed to seekto in compthread %d\n", i);
 
-	if (unlikely(write_i64(cti->sinfo->fd, cti->sinfo->cur_pos)))
+	if (unlikely(write_i64(ctis->fd, ctis->cur_pos)))
 		fatal("Failed to write_i64 in compthread %d\n", i);
 
-	cti->sinfo->s[cti->stream].last_head = cti->sinfo->cur_pos + 17;
-	if (unlikely(seekto(cti->sinfo, cti->sinfo->cur_pos)))
+	ctis->s[cti->stream].last_head = ctis->cur_pos + 17;
+	if (unlikely(seekto(ctis, ctis->cur_pos)))
 		fatal("Failed to seekto cur_pos in compthread %d\n", i);
 
-	print_maxverbose("Writing %lld compressed bytes from thread %ld\n", cti->c_len, i);
-	if (unlikely(write_u8(cti->sinfo->fd, cti->c_type) ||
-		write_i64(cti->sinfo->fd, cti->c_len) ||
-		write_i64(cti->sinfo->fd, cti->s_len) ||
-		write_i64(cti->sinfo->fd, 0))) {
+	print_maxverbose("Thread %ld writing %lld compressed bytes from stream %d\n", i, cti->c_len, cti->stream);
+	if (unlikely(write_u8(ctis->fd, cti->c_type) ||
+		write_i64(ctis->fd, cti->c_len) ||
+		write_i64(ctis->fd, cti->s_len) ||
+		write_i64(ctis->fd, 0))) {
 			fatal("Failed write in compthread %d\n", i);
 	}
-	cti->sinfo->cur_pos += 25;
+	ctis->cur_pos += 25;
 
-	if (unlikely(write_buf(cti->sinfo->fd, cti->s_buf, cti->c_len)))
+	if (unlikely(write_buf(ctis->fd, cti->s_buf, cti->c_len)))
 		fatal("Failed to write_buf in compthread %d\n", i);
 
-	cti->sinfo->cur_pos += cti->c_len;
+	ctis->cur_pos += cti->c_len;
 	free(cti->s_buf);
 
-	fsync(cti->sinfo->fd);
+	fsync(ctis->fd);
 
 	post_sem(&cti->complete);
 	post_sem(&cti->free);
@@ -864,11 +1019,9 @@ void *compthread(void *t)
 	return 0;
 }
 
-/* flush out any data in a stream buffer */
-void flush_buffer(struct stream_info *sinfo, int stream)
+static void clear_buffer(struct stream_info *sinfo, int stream, int newbuf)
 {
-	pthread_t threads[control.threads];
-	long i = sinfo->thread_no;
+	static long i = 0;
 
 	/* Make sure this thread doesn't already exist */
 	wait_sem(&cthread[i].free);
@@ -879,30 +1032,107 @@ void flush_buffer(struct stream_info *sinfo, int stream)
 	cthread[i].s_len = sinfo->s[stream].buflen;
 
 	print_maxverbose("Starting thread %ld to compress %lld bytes from stream %d\n",
-		i, cthread[i].s_len, stream);
+			 i, cthread[i].s_len, stream);
 	create_pthread(&threads[i], NULL, compthread, (void *)i);
 
-	/* The stream buffer has been given to the thread, allocate a new one */
-	sinfo->s[stream].buf = malloc(sinfo->bufsize);
-	if (unlikely(!sinfo->s[stream].buf))
-		fatal("Unable to malloc buffer of size %lld in flush_buffer\n", sinfo->bufsize);
-	sinfo->s[stream].buflen = 0;
+	if (newbuf) {
+		/* The stream buffer has been given to the thread, allocate a new one */
+		sinfo->s[stream].buf = malloc(sinfo->bufsize);
+		if (unlikely(!sinfo->s[stream].buf))
+			fatal("Unable to malloc buffer of size %lld in flush_buffer\n", sinfo->bufsize);
+		sinfo->s[stream].buflen = 0;
+	}
 
-	if (++sinfo->thread_no == control.threads)
-		sinfo->thread_no = 0;
+	if (++i == control.threads)
+		i = 0;
+}
+
+/* flush out any data in a stream buffer */
+void flush_buffer(struct stream_info *sinfo, int stream)
+{
+	clear_buffer(sinfo, stream, 1);
+}
+
+static void *ucompthread(void *t)
+{
+	long i = (long)t;
+	struct uncomp_thread *uci = &ucthread[i];
+	int waited = 0, ret = 0;
+
+	if (unlikely(setpriority(PRIO_PROCESS, 0, control.nice_val) == -1))
+		print_err("Warning, unable to set nice value on thread\n");
+
+retry:
+	if (uci->c_type != CTYPE_NONE) {
+		switch (uci->c_type) {
+			case CTYPE_LZMA:
+				ret = lzma_decompress_buf(uci);
+				break;
+			case CTYPE_LZO:
+				ret = lzo_decompress_buf(uci);
+				break;
+			case CTYPE_BZIP2:
+				ret = bzip2_decompress_buf(uci);
+				break;
+			case CTYPE_GZIP:
+				ret = gzip_decompress_buf(uci);
+				break;
+			case CTYPE_ZPAQ:
+				ret = zpaq_decompress_buf(uci, i);
+				break;
+			default:
+				fatal("Dunno wtf decompression type to use!\n");
+				break;
+		}
+	}
+
+	/* As per compression, serialise the decompression if it fails in
+	 * parallel */
+	if (ret) {
+		if (waited)
+			fatal("Failed to decompress in ucompthread\n");
+		print_maxverbose("Unable to decompress in parallel, waiting for previous thread to complete before trying again\n");
+	}
+
+	/* First ready, are up to this thread yet? */
+	if (!waited)
+		wait_sem(&uci->ready);
+	waited = 1;
+	if (ret)
+		goto retry;
+
+	/* This complete tells the main thread that this thread has its
+	 * decompressed data ready */
+	post_sem(&uci->complete);
+	/* Second ready, the buffer has been grabbed by the main thread so this
+	 * thread can exit */
+	wait_sem(&uci->ready);
+
+	print_maxverbose("Thread %ld returning %lld uncompressed bytes from stream %d\n", i, uci->u_len, uci->stream);
+	post_sem(&uci->free);
+
+	return 0;
 }
 
 /* fill a buffer from a stream - return -1 on failure */
 static int fill_buffer(struct stream_info *sinfo, int stream)
 {
-	i64 header_length, u_len, c_len;
-	uchar c_type;
+	i64 header_length, u_len, c_len, last_head;
+	struct stream *s = &sinfo->s[stream];
+	uchar c_type, *s_buf;
 
-	if (unlikely(seekto(sinfo, sinfo->s[stream].last_head)))
+	if (s->buf)
+		free(s->buf);
+fill_another:
+	if (s->eos)
+		goto out;
+
+	if (unlikely(seekto(sinfo, s->last_head)))
 		return -1;
 
 	if (unlikely(read_u8(sinfo->fd, &c_type)))
 		return -1;
+
 	/* Compatibility crap for versions < 0.4 */
 	if (control.major_version == 0 && control.minor_version < 4) {
 		u32 c_len32, u_len32, last_head32;
@@ -915,53 +1145,70 @@ static int fill_buffer(struct stream_info *sinfo, int stream)
 			return -1;
 		c_len = c_len32;
 		u_len = u_len32;
-		sinfo->s[stream].last_head = last_head32;
+		last_head = last_head32;
 		header_length = 13;
 	} else {
 		if (unlikely(read_i64(sinfo->fd, &c_len)))
 			return -1;
 		if (unlikely(read_i64(sinfo->fd, &u_len)))
 			return -1;
-		if (unlikely(read_i64(sinfo->fd, &sinfo->s[stream].last_head)))
+		if (unlikely(read_i64(sinfo->fd, &last_head)))
 			return -1;
 		header_length = 25;
 	}
 
 	sinfo->total_read += header_length;
 
-	if (sinfo->s[stream].buf)
-		sinfo->s[stream].buf = realloc(sinfo->s[stream].buf, u_len);
-	else
-		sinfo->s[stream].buf = malloc(u_len);
-	if (unlikely(u_len && !sinfo->s[stream].buf))
+	/* Wait till the next thread is free */
+	wait_sem(&ucthread[s->uthread_no].free);
+
+	s_buf = malloc(u_len);
+	if (unlikely(u_len && !s_buf))
 		fatal("Unable to malloc buffer of size %lld in fill_buffer\n", u_len);
 
-	if (unlikely(read_buf(sinfo->fd, sinfo->s[stream].buf, c_len)))
+	if (unlikely(read_buf(sinfo->fd, s_buf, c_len)))
 		return -1;
 
 	sinfo->total_read += c_len;
 
-	sinfo->s[stream].buflen = u_len;
-	sinfo->s[stream].bufp = 0;
+	ucthread[s->uthread_no].s_buf = s_buf;
+	ucthread[s->uthread_no].c_len = c_len;
+	ucthread[s->uthread_no].u_len = u_len;
+	ucthread[s->uthread_no].c_type = c_type;
+	ucthread[s->uthread_no].stream = stream;
+	s->last_head = last_head;
 
-	if (c_type != CTYPE_NONE) {
-		if (c_type == CTYPE_LZMA) {
-			if (unlikely(lzma_decompress_buf(&sinfo->s[stream], (size_t)c_len)))
-				return -1;
-		} else if (c_type == CTYPE_LZO) {
-			if (unlikely(lzo_decompress_buf(&sinfo->s[stream], c_len)))
-				return -1;
-		} else if (c_type == CTYPE_BZIP2) {
-			if (unlikely(bzip2_decompress_buf(&sinfo->s[stream], c_len)))
-				return -1;
-		} else if (c_type == CTYPE_GZIP) {
-			if (unlikely(gzip_decompress_buf(&sinfo->s[stream], c_len)))
-				return -1;
-		} else if (c_type == CTYPE_ZPAQ) {
-			if (unlikely(zpaq_decompress_buf(&sinfo->s[stream])))
-				return -1;
-		} else fatal("Dunno wtf decompression type to use!\n");
+	print_maxverbose("Starting thread %ld to decompress %lld bytes from stream %d\n",
+			 s->uthread_no, c_len, stream);
+	create_pthread(&threads[s->uthread_no], NULL, ucompthread, (void *)s->uthread_no);
+
+	if (!last_head)
+		s->eos = 1;
+
+	if (++s->uthread_no == s->base_thread + s->total_threads)
+		s->uthread_no = s->base_thread;
+	if (s->uthread_no != s->unext_thread &&
+	    !trywait_sem(&ucthread[s->uthread_no].free)) {
+		post_sem(&ucthread[s->uthread_no].free);
+		goto fill_another;
 	}
+out:
+	/* This ready tells the decompression thread we're up to it */
+	post_sem(&ucthread[s->unext_thread].ready);
+	/* This complete is the deco thread telling us it's finished
+	 * decompressing data */
+	wait_sem(&ucthread[s->unext_thread].complete);
+
+	s->buf = ucthread[s->unext_thread].s_buf;
+	s->buflen = ucthread[s->unext_thread].u_len;
+	s->bufp = 0;
+
+	/* This ready tells the decompression thread we've taken the buffer so
+	 * it can exit */
+	post_sem(&ucthread[s->unext_thread].ready);
+
+	if (++s->unext_thread == s->base_thread + s->total_threads)
+		s->unext_thread = s->base_thread;
 
 	return 0;
 }
@@ -1027,22 +1274,16 @@ int close_stream_out(void *ss)
 
 	for (i = 0; i < sinfo->num_streams; i++) {
 		if (sinfo->s[i].buflen)
-			flush_buffer(sinfo, i);
+			clear_buffer(sinfo, i, 0);
 	}
 
-	for (i = 0; i < control.threads; i++) {
-		wait_sem(&cthread[i].free);
-		destroy_sem(&cthread[i].complete);
-		destroy_sem(&cthread[i].free);
-	}
-
-	for (i = 0; i < sinfo->num_streams; i++)
-		free(sinfo->s[i].buf);
-
-	free(cthread);
+#if 0
+	/* These cannot be freed because their values are read after the next
+	 * stream has started so they're not properly freed and just dropped on
+	 * program exit! FIXME */
 	free(sinfo->s);
 	free(sinfo);
-
+#endif
 	return 0;
 }
 
@@ -1058,8 +1299,18 @@ int close_stream_in(void *ss)
 	for (i = 0; i < sinfo->num_streams; i++)
 		free(sinfo->s[i].buf);
 
+	for (i = 0; i < control.threads + 1; i++) {
+		wait_sem(&ucthread[i].free);
+		destroy_sem(&ucthread[i].complete);
+		destroy_sem(&ucthread[i].free);
+		destroy_sem(&ucthread[i].ready);
+	}
+
+	free(ucthread);
+	free(threads);
 	free(sinfo->s);
 	free(sinfo);
+
 	return 0;
 }
 
@@ -1093,7 +1344,7 @@ static int lzo_compresses(uchar *s_buf, i64 s_len)
 	if (unlikely(!c_buf))
 		fatal("Unable to allocate c_buf in lzo_compresses\n");
 
-	print_progress("lzo testing for incompressible data...\n");
+	print_verbose("lzo testing for incompressible data...\n");
 
 	/* Test progressively larger blocks at a time and as soon as anything
 	   compressible is found, jump out as a success */
